import abc


class BaseMemory(abc.ABC):
    @abc.abstractmethod
    def inc_call(self, key: str, ttl: float): ...

    @abc.abstractmethod
    def call_of(self, key: str) -> int: ...

    @abc.abstractmethod
    def exists(self, key: str) -> bool: ...



from typing import Callable, Union, ClassVar, Type
import functools
import uuid
from .memory_abc import BaseMemory
from .expiringdict_memory import ExpiringDictMemory

KeyType = Callable[[tuple, dict], str]
ExceptionType = Union[Type[BaseException], BaseException]


class BaseLimiter:
    times: int
    per: float
    key: KeyType = lambda *args, **kwargs: ''
    exception: ExceptionType
    memory: ClassVar[BaseMemory] = ExpiringDictMemory()

    def __init__(self, times: int, per: float, key: KeyType = lambda *args, **kwargs: '',
                 exception: ExceptionType = OverflowError):
        self.times = times
        self.per = per
        self.key = key
        self.exception = exception

    def __call__(self, func):
        base_key = str(uuid.uuid4())

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            key = base_key + self.key(*args, **kwargs)
            if self.memory.call_of(key) >= self.times:
                raise self.exception
            self.memory.inc_call(key, self.per)
            return func(*args, **kwargs)

        return wrapper



from typing import ClassVar


class Per:
    MILLISECOND: ClassVar[float] = .001
    SECOND: ClassVar[int] = 1
    MINUTE: ClassVar[int] = 60
    HOUR: ClassVar[int] = 60 * 60
    DAY: ClassVar[int] = 24 * 60 * 60



from typing import Dict, List
from time import time
from .memory_abc import BaseMemory


class ExpiringDictMemory(BaseMemory):
    def __init__(self):
        self.data: Dict[str, List[float]] = dict()

    def inc_call(self, key: str, ttl: float):
        if not self.exists(key):
            self.data[key] = list()
        self.data[key].append(time() + ttl)

    def call_of(self, key: str) -> int:
        if not self.exists(key):
            return 0
        now, less_than_now_index, current_length = time(), 0, len(self.data[key])
        while less_than_now_index < current_length \
            and self.data[key][less_than_now_index] <= now:
            less_than_now_index += 1
        active_calls = len(self.data[key][less_than_now_index:])
        del self.data[key][:less_than_now_index]
        return active_calls

    def exists(self, key: str) -> bool:
        return key in self.data



from rlimit.per import Per
from rlimit.memory_abc import BaseMemory
from rlimit.expiringdict_memory import ExpiringDictMemory
from rlimit.base_limiter import BaseLimiter
from rlimit.base_limiter import BaseLimiter as Limiter
